
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../libmad/mad.h"
#include "audio.h"
#include "player.h"
#include "../mp3_18.h"

# define MPEG_BUFSZ	40000	/* 2.5 s at 128 kbps; 1 s at 320 kbps */
# define FREQ_TOLERANCE	2	/* percent sampling frequency tolerance */

//static int player_run(struct player *player, TCHAR *playfile);
//static void player_init(struct player *player);
//static void player_finish(struct player *player);
static int play_one(struct player *player);
static int decode(struct player *player);
static enum mad_flow decode_input_read(void *data, struct mad_stream *stream);
static enum mad_flow decode_header(void *data, struct mad_header const *header);
static enum mad_flow decode_output(void *data, struct mad_header const *header, struct mad_pcm *pcm);
static enum mad_flow decode_error(void *data, struct mad_stream *stream, struct mad_frame *frame);
static enum mad_flow decode_filter(void *data, struct mad_stream const *stream, struct mad_frame *frame);
static int setup_filters(struct player *player);
static double set_gain(struct player *player, int how, double db);
static void error(char const *id, char const *format, ...);
static int message(char const *format, ...);

void start_audio_i2s(unsigned int sample_freq, unsigned short channel_count, unsigned short sample_stride);		// audio_i2s.c

enum {
	GAIN_VOLADJ   = 0x0001,
	GAIN_ATTAMP   = 0x0002,
	GAIN_RELATIVE = 0x0010
};

extern int 				EventSw;	// (main.c)
extern uint8_t			RandomPlaying;	// 1=ランダム再生中 (main.c)
extern volatile uint32_t	AutoPoffTimer;		// 自動停止タイマーカウンタ


/*
* NAME:	player_init()
* DESCRIPTION:	initialize player structure
*/
//static
void player_init(struct player *player)
{
	player->verbosity = 1;		//23/5/27 0;
	
	player->options = 0;
	player->repeat  = 1;
	
	player->control = PLAYER_CONTROL_DEFAULT;
	
	player->playfile = 0;			// 23/5/27
#if 0		// 23/5/27
	player->playlist.entries = 0;
	player->playlist.length  = 0;
	player->playlist.current = 0;
#endif

	player->global_start = mad_timer_zero;
	player->global_stop  = mad_timer_zero;
	
	player->fade_in      = mad_timer_zero;
	player->fade_out     = mad_timer_zero;
	player->gap          = mad_timer_zero;
	
//	player->input.path   = 0;
//	player->input.fd     = -1;
# if defined(HAVE_MMAP)
	player->input.fdm    = 0;
# endif
	player->input.data   = 0;
	player->input.length = 0;
	player->input.eof    = 0;
	
//23/5/27	tag_init(&player->input.tag);
	
	player->output.mode          = AUDIO_MODE_DITHER;
	player->output.voladj_db     = 0;
	player->output.attamp_db     = 0;
	player->output.gain          = MAD_F_ONE;
	player->output.replay_gain   = 0;
	player->output.filters       = 0;
	player->output.channels_in   = 0;
	player->output.channels_out  = 0;
	player->output.select        = PLAYER_CHANNEL_DEFAULT;
	player->output.speed_in      = 0;
	player->output.speed_out     = 0;
	player->output.speed_request = 0;
	player->output.precision_in  = 0;
	player->output.precision_out = 0;
	//player->output.path          = 0;
	//player->output.command       = 0;
	player->output.command = audio_wave;		// audio_output(&player.output.path);
	/* player->output.resample */
	player->output.resampled     = 0;
	player->output.init_i2s      = 0;		// 2023/07/29
	
#if 0		//2023/5/27
	player->ancillary.path       = 0;
	player->ancillary.file       = 0;
	player->ancillary.buffer     = 0;
	player->ancillary.length     = 0;
#endif

#if 0		// 2023/11/11
	player->stats.show                  = STATS_SHOW_OVERALL;
	player->stats.label                 = 0;
	player->stats.total_bytes           = 0;
	player->stats.total_time            = mad_timer_zero;
	player->stats.global_timer          = mad_timer_zero;
	player->stats.absolute_timer        = mad_timer_zero;
	player->stats.play_timer            = mad_timer_zero;
#endif
	player->stats.global_framecount     = 0;
	player->stats.absolute_framecount   = 0;
	player->stats.play_framecount       = 0;
	player->stats.error_frame           = -1;
	player->stats.mute_frame            = 0;
	player->stats.vbr                   = 0;
	player->stats.bitrate               = 0;
	player->stats.vbr_frames            = 0;
	player->stats.vbr_rate              = 0;
	player->stats.nsecs                 = 0;
	player->stats.audio.clipped_samples = 0;
	player->stats.audio.peak_clipping   = 0;
	player->stats.audio.peak_sample     = 0;
}

/*
* NAME:	player->run()
* DESCRIPTION:	begin playback
*/
//static
int player_run(struct player *player, TCHAR *playfile, double gain)
{
	int result = 0;
	union audio_control control;
	
	player->playfile = playfile;

	/* set up filters */
	
	if (setup_filters(player) == -1) {
		error("filter", "not enough memory to allocate filters");
		goto fail;
	}
	
	//set_gain(player, 0, 0);
	//set_gain(player, GAIN_ATTAMP, -30);		// 2023/07/29
	set_gain(player, GAIN_ATTAMP, gain);		// 2023/07/29
	
	/* initialize audio */
	if (player->output.command) {		// audio_wave() (madplay\audio_wave.c)
		audio_control_init(&control, AUDIO_COMMAND_INIT);
		//control.init.path = player->output.path;		// 2023/07/30
		
		if (player->output.command(&control) == -1) {			// init() (madplay\audio_wave.c)
			error("audio", audio_error, control.init.path);
			goto fail;
		}
	}
	
	if (result == 0){
		player->control = PLAYER_CONTROL_DEFAULT;
		if ((result = decode(player)) == -1) {	// play_one -> decode
			result = -1;
		}
	}
	
	/* drain and close audio */
	
	if (player->output.command) {
		audio_control_init(&control, AUDIO_COMMAND_FINISH);
		
		if (player->output.command(&control) == -1) {
			error("audio", audio_error);
			goto fail;
		}
	}
	
	if (0) {
fail:
    result = -1;
	}
	
	return result;
}


/*
* NAME:	player_finish()
* DESCRIPTION:	destroy a player structure
*/
//static
void player_finish(struct player *player)
{
	if (player->output.filters)
		filter_free(player->output.filters);

	if (player->output.resampled) {
		resample_finish(&player->output.resample[0]);
		resample_finish(&player->output.resample[1]);
		
		free(player->output.resampled);
		player->output.resampled = 0;
	}
}


/*
* NAME:	play_one()
* DESCRIPTION:	open and play a single file
*/
static
int play_one(struct player *player)
{
	TCHAR const* file = player->playfile;
	int result;

//	player->input.path = file;
	
	/* reset file information */
	
//	player->stats.total_bytes = 0;
//	player->stats.total_time  = mad_timer_zero;
	
	result = decode(player);
	
	return result;
}

/*
* NAME:	decode()
* DESCRIPTION:	decode and output audio for an open file
*/
static
int decode(struct player *player)
{
	//struct stat stat;
	struct mad_decoder decoder;
	int options, result;

//	player->stats.total_bytes = f_size(&player->input.fd);

	if (player->input.data == 0) {
		player->input.data = malloc(MPEG_BUFSZ);
		if (player->input.data == 0) {
			error("decode", "not enough memory to allocate input buffer");
			return -1;
		}
		
		player->input.length = 0;
	}
	
	player->input.eof = 0;
	
	/* reset statistics */
#if 0	// 2023/11/11
	player->stats.absolute_timer        = mad_timer_zero;
	player->stats.play_timer            = mad_timer_zero;
#endif
	player->stats.absolute_framecount   = 0;
	player->stats.play_framecount       = 0;
	player->stats.error_frame           = -1;
	player->stats.vbr                   = 0;
	player->stats.bitrate               = 0;
	player->stats.vbr_frames            = 0;
	player->stats.vbr_rate              = 0;
	player->stats.audio.clipped_samples = 0;
	player->stats.audio.peak_clipping   = 0;
	player->stats.audio.peak_sample     = 0;
	
	mad_decoder_init(&decoder, player,
		decode_input_read,
		decode_header, decode_filter,
		player->output.command ? decode_output : 0,
		decode_error, 0);
		
	options = 0;

	mad_decoder_options(&decoder, options);
	
	result = mad_decoder_run_sync(&decoder);		// decoder.c
	
	mad_decoder_finish(&decoder);
	
	if (player->input.data) {
		free(player->input.data);
		player->input.data = 0;
	}
	
	return result;
}


/*
* NAME:	decode->input_read()
* DESCRIPTION:	(re)fill decoder input buffer by reading a file descriptor
*/
static
enum mad_flow decode_input_read(void *data, struct mad_stream *stream)
{
	struct player *player = data;
	struct input *input = &player->input;
	int len;
    FRESULT fr;     /* FatFs return code */
	
	//printf("decode_input_read\n");
	
	if (input->eof)
		return MAD_FLOW_STOP;
	
	if (stream->next_frame) {
		memmove(input->data, stream->next_frame,
			input->length = &input->data[input->length] - stream->next_frame);
	}
	
	fr = f_read(&input->fd, input->data + input->length, MPEG_BUFSZ - input->length, &len);
//	do {
//		len = read(input->fd, input->data + input->length, MPEG_BUFSZ - input->length);
//	}
//	while (len == -1 && errno == EINTR);
	
	if (fr != FR_OK){
	//if (len == -1) {
		error("input", ":read");
		return MAD_FLOW_BREAK;
	}
	else if (len == 0) {
		input->eof = 1;
		
		//assert(MPEG_BUFSZ - input->length >= MAD_BUFFER_GUARD);
		
		while (len < MAD_BUFFER_GUARD)
			input->data[input->length + len++] = 0;
	}
	
	mad_stream_buffer(stream, input->data, input->length += len);
	
	return MAD_FLOW_CONTINUE;
}


/*
* NAME:	decode->header()
* DESCRIPTION:	decide whether to continue decoding based on header
*/
static
enum mad_flow decode_header(void *data, struct mad_header const *header)
{
	struct player *player = data;
	
	if (player->output.init_i2s == 0){
		player->output.init_i2s = 1;
		printf("header %d samplerate=%d mode=%d\n", player->stats.play_framecount, header->samplerate, header->mode);		// 2023/07/29
		if (header->mode == MAD_MODE_SINGLE_CHANNEL){
			player->output.select = PLAYER_CHANNEL_STEREO;
		}
		//unsigned short ch = (header->mode == MAD_MODE_SINGLE_CHANNEL ? 1 : 2);		// libmad/frame.h
		//unsigned short stride = (ch == 1 ? 2 : 4);
		//start_audio_i2s(header->samplerate, ch, stride);
		start_audio_i2s(header->samplerate, 2, 4);
	}
	
//	if ((player->options & PLAYER_OPTION_TIMED) &&
//		mad_timer_compare(player->stats.global_timer, player->global_stop) > 0)
//		return MAD_FLOW_STOP;
	
	/* accounting (except first frame) */
	
	if (player->stats.absolute_framecount) {
		++player->stats.absolute_framecount;
//		mad_timer_add(&player->stats.absolute_timer, header->duration);
		
		++player->stats.global_framecount;
//		mad_timer_add(&player->stats.global_timer, header->duration);
		
//		if ((player->options & PLAYER_OPTION_SKIP) &&
//			mad_timer_compare(player->stats.global_timer,
//			player->global_start) < 0)
//			return MAD_FLOW_IGNORE;
	}
	
	return MAD_FLOW_CONTINUE;
}


/*
* NAME:	decode->output()
* DESCRIPTION: configure audio module and output decoded samples
*/
static
enum mad_flow decode_output(void *data, struct mad_header const *header,
							struct mad_pcm *pcm)
{
	struct player *player = data;
	struct output *output = &player->output;
	mad_fixed_t const *ch1, *ch2;
	unsigned int nchannels;
	union audio_control control;
	
	//printf("decode_output\n");
	
	ch1 = pcm->samples[0];
	ch2 = pcm->samples[1];
	
	switch (nchannels = pcm->channels) {
	case 1:
		ch2 = 0;
		if (output->select == PLAYER_CHANNEL_STEREO) {
			ch2 = ch1;
			nchannels = 2;
		}
		break;
		
	case 2:
		switch (output->select) {
		case PLAYER_CHANNEL_RIGHT:
			ch1 = ch2;
			/* fall through */
			
		case PLAYER_CHANNEL_LEFT:
			ch2 = 0;
			nchannels = 1;
			/* fall through */
			
		case PLAYER_CHANNEL_STEREO:
			break;
			
		default:
			if (header->mode == MAD_MODE_DUAL_CHANNEL) {
				if (output->select == PLAYER_CHANNEL_DEFAULT) {
					if (player->verbosity >= -1) {
						error("output",
							"no channel selected for dual channel; using first");
					}
					
					output->select = -PLAYER_CHANNEL_LEFT;
				}
				
				ch2 = 0;
				nchannels = 1;
			}
		}
	}
	
	if (output->channels_in != nchannels ||
		output->speed_in != pcm->samplerate) {
		unsigned int speed_request;
		
		printf("first output\n");
		if (player->verbosity >= 1 &&
			pcm->samplerate != header->samplerate) {
			error("output", "decoded sample frequency %u Hz",
				pcm->samplerate);
		}
		
		speed_request = output->speed_request ?
			output->speed_request : pcm->samplerate;
		
		audio_control_init(&control, AUDIO_COMMAND_CONFIG);
		
		control.config.channels  = nchannels;
		control.config.speed     = speed_request;
		control.config.precision = output->precision_in;
		
		if (output->command(&control) == -1) {
			error("output", audio_error);
			return MAD_FLOW_BREAK;
		}
		
		output->channels_in  = nchannels;
		output->speed_in     = pcm->samplerate;
		
		output->channels_out  = control.config.channels;
		output->speed_out     = control.config.speed;
		output->precision_out = control.config.precision;
		
		if (player->verbosity >= -1 &&
			output->channels_in != output->channels_out) {
			if (output->channels_in == 1)
				error("output", "mono output not available; forcing stereo");
			else {
				error("output", "stereo output not available; using first channel "
					"(use -m to mix)");
			}
		}
		
		if (player->verbosity >= -1 &&
			output->precision_in &&
			output->precision_in != output->precision_out) {
			error("output", "bit depth %u not available; using %u",
				output->precision_in, output->precision_out);
		}
		
		if (player->verbosity >= -1 &&
			speed_request != output->speed_out) {
			error("output", "sample frequency %u Hz not available; using %u Hz",
				speed_request, output->speed_out);
		}
		
		/* check whether resampling is necessary */
		if (abs(output->speed_out - output->speed_in) <
			(long) FREQ_TOLERANCE * output->speed_in / 100) {
			//printf("resampling NO need\n");
			if (output->resampled) {
				resample_finish(&output->resample[0]);
				resample_finish(&output->resample[1]);
				
				free(output->resampled);
				output->resampled = 0;
			}
		}
		else {
			//printf("resampling NEED\n");
			if (output->resampled) {
				resample_finish(&output->resample[0]);
				resample_finish(&output->resample[1]);
			}
			else {
				output->resampled = malloc(sizeof(*output->resampled));
				if (output->resampled == 0) {
					error("output",
						"not enough memory to allocate resampling buffer");
					
					output->speed_in = 0;
					return MAD_FLOW_BREAK;
				}
			}
			
			if (resample_init(&output->resample[0],
				output->speed_in, output->speed_out) == -1 ||
				resample_init(&output->resample[1],
				output->speed_in, output->speed_out) == -1) {
				error("output", "cannot resample %u Hz to %u Hz",
					output->speed_in, output->speed_out);
				
				free(output->resampled);
				output->resampled = 0;
				
				output->speed_in = 0;
				return MAD_FLOW_BREAK;
			}
			else if (player->verbosity >= -1) {
				error("output", "resampling %u Hz to %u Hz",
					output->speed_in, output->speed_out);
			}
		}
	}
	
		audio_control_init(&control, AUDIO_COMMAND_PLAY);		// madplay\audio.c
		
		if (output->channels_in != output->channels_out)
			ch2 = (output->channels_out == 2) ? ch1 : 0;
		
		if (output->resampled) {
			//printf("resampled\n");
			control.play.nsamples = resample_block(&output->resample[0],
				pcm->length, ch1,
				(*output->resampled)[0]);
			control.play.samples[0] = (*output->resampled)[0];
			
			if (ch2 == ch1)
				control.play.samples[1] = control.play.samples[0];
			else if (ch2) {
				resample_block(&output->resample[1], pcm->length, ch2,
					(*output->resampled)[1]);
				control.play.samples[1] = (*output->resampled)[1];
			}
			else
				control.play.samples[1] = 0;
		}
		else {
			control.play.nsamples   = pcm->length;
			control.play.samples[0] = ch1;
			control.play.samples[1] = ch2;
		}
		
		control.play.mode  = output->mode;
		control.play.stats = &player->stats.audio;
		
		if (output->command(&control) == -1) {		// play() (madplay\audio_wave.c)
			error("output", audio_error);
			return MAD_FLOW_BREAK;
		}
		
		++player->stats.play_framecount;
//		mad_timer_add(&player->stats.play_timer, header->duration);
		
#if 0		// 23/5/29
		if (player->verbosity > 0)
			show_status(&player->stats, header, 0, 0);
#endif
		
		return MAD_FLOW_CONTINUE;
}


/*
* NAME:	decode->error()
* DESCRIPTION:	handle a decoding error
*/
static
enum mad_flow decode_error(void *data, struct mad_stream *stream,
						   struct mad_frame *frame)
{
	struct player *player = data;
	signed long tagsize;
	
	switch (stream->error) {
	case MAD_ERROR_BADDATAPTR:
		return MAD_FLOW_CONTINUE;
		
	case MAD_ERROR_LOSTSYNC:
#if 0		// 23/5/27
		tagsize = id3_tag_query(stream->this_frame,
			stream->bufend - stream->this_frame);
		if (tagsize > 0) {
			if (player->options & PLAYER_OPTION_STREAMID3) {
				struct id3_tag *tag;
				
				tag = get_id3(stream, tagsize, &player->input);
				if (tag) {
					process_id3(tag, player);
					id3_tag_delete(tag);
				}
			}
			else
				mad_stream_skip(stream, tagsize);
			
			if (player->stats.total_bytes >= tagsize)
				player->stats.total_bytes -= tagsize;
			
			return MAD_FLOW_CONTINUE;
		}
#endif

		/* fall through */
		
	default:
		if (player->verbosity >= -1 &&
			!(player->options & PLAYER_OPTION_SHOWTAGSONLY) &&
			((stream->error == MAD_ERROR_LOSTSYNC && !player->input.eof)
			|| stream->sync) &&
			player->stats.global_framecount != player->stats.error_frame) {
			error("error", "frame %lu: %s",
				player->stats.absolute_framecount, mad_stream_errorstr(stream));
			player->stats.error_frame = player->stats.global_framecount;
		}
	}
	
	if (stream->error == MAD_ERROR_BADCRC) {
		if (player->stats.global_framecount == player->stats.mute_frame)
			mad_frame_mute(frame);
		
		player->stats.mute_frame = player->stats.global_framecount + 1;
		
		return MAD_FLOW_IGNORE;
	}
	
	return MAD_FLOW_CONTINUE;
}


/* 2023/07/29
* NAME:	decode->filter()
* DESCRIPTION:	perform filtering on decoded frame
* MAD_FLOW_***定数はdecoder.hで定義
*/
static
enum mad_flow decode_filter(void *data, struct mad_stream const *stream,
							struct mad_frame *frame)
{
	struct player *player = data;
	
	if (AutoPoffTimer > AUTOPOFF_PLAYING){
		EventSw = -1;
		return MAD_FLOW_POWEROFF;
	}
	
	if (EventSw == PIN_SW_UP){
		EventSw = -1;
		//printf("UP\n");
		set_gain(player, GAIN_ATTAMP | GAIN_RELATIVE, +1);
	} else if (EventSw == PIN_SW_PGUP || EventSw == PIN_RE_UP){
		EventSw = -1;
		//printf("PGUP\n");
		set_gain(player, GAIN_ATTAMP | GAIN_RELATIVE, +4);
	} else if (EventSw == PIN_SW_DOWN){
		EventSw = -1;
		//printf("DOWN\n");
		set_gain(player, GAIN_ATTAMP | GAIN_RELATIVE, -1);
	} else if (EventSw == PIN_SW_PGDOWN || EventSw == PIN_RE_DOWN){
		EventSw = -1;
		//printf("PGDOWN\n");
		set_gain(player, GAIN_ATTAMP | GAIN_RELATIVE, -4);
	} else if (EventSw == PIN_SW_RANDOM){
		EventSw = -1;
		if (RandomPlaying == 1) return MAD_FLOW_RANDOM;		// 再生中にRANDOMボタンが使えるのはランダム再生中だけ
	} else if (EventSw == PIN_SW_PLAY){
		EventSw = -1;
		return MAD_FLOW_PLAY;
	} else if (EventSw == PIN_SW_RETURN){
		EventSw = -1;
		return MAD_FLOW_RETURN;
	} else if (EventSw == PIN_SW_POWER_OFF){
		EventSw = -1;
		return MAD_FLOW_POWEROFF;
	} 
	
#if 0		// 23/07/29
	/* output ancillary data */
	
	if (player->ancillary.file && stream->anc_bitlen &&
		write_ancillary(&player->ancillary,
		      stream->anc_ptr, stream->anc_bitlen) == -1)
			  return MAD_FLOW_BREAK;
#endif
	
	/* first frame accounting */
	
	if (player->stats.absolute_framecount == 0) {
#if 0	// 23/07/29
		if (player->input.tag.flags == 0 &&
			tag_parse(&player->input.tag, stream) == 0) {
			struct tag *tag = &player->input.tag;
			unsigned int frame_size;
			
			if (player->options & PLAYER_OPTION_SHOWTAGSONLY) {
				if (player->verbosity > 0)
					show_tag(tag);
			}
			else {
				if ((tag->flags & TAG_LAME) &&
					(player->output.replay_gain & PLAYER_RGAIN_ENABLED) &&
					!(player->output.replay_gain & PLAYER_RGAIN_SET))
					use_rgain(player, tag->lame.replay_gain);
			}
			
			if ((tag->flags & TAG_XING) &&
				(tag->xing.flags & TAG_XING_FRAMES)) {
				player->stats.total_time = frame->header.duration;
				mad_timer_multiply(&player->stats.total_time, tag->xing.frames);
			}
			
			/* total stream byte size adjustment */
			
			frame_size = stream->next_frame - stream->this_frame;
			
			if (player->stats.total_bytes == 0) {
				if ((tag->flags & TAG_XING) && (tag->xing.flags & TAG_XING_BYTES) &&
					tag->xing.bytes > frame_size)
					player->stats.total_bytes = tag->xing.bytes - frame_size;
			}
			else if (player->stats.total_bytes >=
				stream->next_frame - stream->this_frame)
				player->stats.total_bytes -= frame_size;
			
			return (player->options & PLAYER_OPTION_SHOWTAGSONLY) ?
MAD_FLOW_STOP : MAD_FLOW_IGNORE;
		}
		else if (player->options & PLAYER_OPTION_SHOWTAGSONLY)
			return MAD_FLOW_STOP;
#endif
		
		++player->stats.absolute_framecount;
//		mad_timer_add(&player->stats.absolute_timer, frame->header.duration);
		
		++player->stats.global_framecount;
//		mad_timer_add(&player->stats.global_timer, frame->header.duration);
		
//		if ((player->options & PLAYER_OPTION_SKIP) &&
//			mad_timer_compare(player->stats.global_timer,
//			player->global_start) < 0)
//			return MAD_FLOW_IGNORE;
	}
	
	/* run the filter chain */
	
	return filter_run(player->output.filters, frame);
}

/*
* NAME:	addfilter()
* DESCRIPTION:	insert a filter at the beginning of the filter chain
*/
static
int addfilter(struct player *player, filter_func_t *func, void *data)
{
	struct filter *filter;
	
	filter = filter_new(func, data, player->output.filters);
	if (filter == 0)
		return -1;
	
	player->output.filters = filter;
	
	return 0;
}

/*
* NAME:	setup_filters()
* DESCRIPTION:	create output filters
*/
static
int setup_filters(struct player *player)
{
	/* filters must be added in reverse order */
	
# if defined(EXPERIMENTAL)
	if ((player->options & PLAYER_OPTION_EXTERNALMIX) &&
		addfilter(player, mixer_filter, stdout) == -1)
		return -1;
	
	if ((player->options & PLAYER_OPTION_EXPERIMENTAL) &&
		addfilter(player, experimental_filter, 0) == -1)
		return -1;
# endif
	
#if 0		// 2023/11/11
	if ((player->options & PLAYER_OPTION_FADEIN) &&
		addfilter(player, fadein_filter, player) == -1)
		return -1;
#endif
	
	addfilter(player, gain_filter, &player->output.gain);
	
	if (player->output.select == PLAYER_CHANNEL_MONO &&
		addfilter(player, mono_filter, player) == -1)
		return -1;
	
# if defined(USE_TTY)
	if ((player->options & PLAYER_OPTION_TTYCONTROL) &&
		addfilter(player, tty_filter, player) == -1)
		return -1;
# endif
	
	return 0;
}



/*
* NAME:	set_gain()
* DESCRIPTION:	modify player gain information
*/
static
double set_gain(struct player *player, int how, double db)
{
	double *gain_db = 0;
	
	if (how & GAIN_ATTAMP)
		gain_db = &player->output.attamp_db;
	else if (how & GAIN_VOLADJ)
		gain_db = &player->output.voladj_db;
	
	if (gain_db) {
		if (how & GAIN_RELATIVE)
			*gain_db += db;
		else
			*gain_db  = db;
	}
	
	db = player->output.voladj_db + player->output.attamp_db;
	if (db > DB_MAX || db < DB_MIN) {
		db = (db > DB_MAX) ? DB_MAX : DB_MIN;
		player->output.attamp_db = db - player->output.voladj_db;
	}
	
	player->output.gain = db ? mad_f_tofixed(pow(10, db / 20)) : MAD_F_ONE;
	
	return db;
}


/*
* NAME:	error()
* DESCRIPTION:	show an error using proper interaction with message()
*/
static
void error(char const *id, char const *format, ...)
{
    va_list va;
    va_start(va, format);
    printf("%s : ", id);
    int ret = vprintf(format, va);
    va_end(va);
    printf("\n");
}

/*
* NAME:	message()
* DESCRIPTION:	show a message, possibly overwriting a previous w/o newline
*/
static
int message(char const *format, ...)
{
    va_list va;
    va_start(va, format);
    int ret = vprintf(format, va);
    va_end(va);
    printf("\n");
	return 0;
}

