//#include <cstdint>
//#include <cstdio>
#include <stdio.h>
#include <string.h>

#include "pico/stdlib.h"
#include "hardware/i2c.h"
#include "hardware/uart.h"
#include "fatfs/ff.h"
#include "pico/audio_i2s.h"
#include "madplay/player.h"
#include "mp3_18.h"

#if PICO_ON_DEVICE
#include "pico/binary_info.h"
bi_decl(bi_3pins_with_names(PICO_AUDIO_I2S_DATA_PIN, "I2S DIN", PICO_AUDIO_I2S_CLOCK_PIN_BASE, "I2S BCK", PICO_AUDIO_I2S_CLOCK_PIN_BASE+1, "I2S LRCK"));
#endif

#if !defined(TRUE)
	#define		TRUE			1
#endif
#if !defined(FALSE)
	#define		FALSE			0
#endif

// ボタン操作(PIN_SW_***はGPの番号。GP0を使う予定はないが+1しておく)
#define SWSELECT_PRESSED      	(PIN_SW_SELECT + 1)
#define SWRETURN_PRESSED      	(PIN_SW_RETURN + 1)
#define SWDOWN_PRESSED          (PIN_SW_DOWN + 1)
#define SWUP_PRESSED            (PIN_SW_UP + 1)
#define SWOPTION_PRESSED     	(PIN_SW_OPTION + 1)
#define SWPLAY_PRESSED          (PIN_SW_PLAY + 1)
#define SWRANDOM_PRESSED        (PIN_SW_RANDOM + 1)
#define SWPAGEDOWN_PRESSED      (PIN_SW_PGDOWN + 1)
#define SWPAGEUP_PRESSED        (PIN_SW_PGUP + 1)
#define SWPOFF_PRESSED			(PIN_SW_POWER_OFF + 1)
#define	REDOWN_PRESSED			(PIN_RE_DOWN + 1)
#define	REUP_PRESSED			(PIN_RE_UP + 1)

// DIR構成ファイル　レコード長・項目別オフセット＆バイトサイズ
#define	DIRSTRUCTUREFILE				"/DIRLST18.DAT"		// DIR構成ファイル(Version=18)
#define	RECLEN_DIRLIST					512UL				// イメージ部分を除くレコード長
#define	RECLEN_DIRLIST_FULL				(512UL + 512UL)		// イメージ部分を含むレコード長
#define DL_DISPLAYNAME_OFFSET			0					// ディレクトリ表示名
#define DL_DISPLAYNAME_LEN				32
#define DL_FULLDIR_OFFSET				(DL_DISPLAYNAME_OFFSET + DL_DISPLAYNAME_LEN)	// ディレクトリフルパス
#define DL_FULLDIR_LEN					258
#define DL_FULLDIRBYTES_OFFSET			(DL_FULLDIR_OFFSET + DL_FULLDIR_LEN)			// ディレクトリフルパスバイト数
#define DL_FULLDIRBYTES_LEN				2
#define DL_DIRNAME_OFFSET				(DL_FULLDIRBYTES_OFFSET + DL_FULLDIRBYTES_LEN)	// ディレクトリ名
#define DL_DIRNAME_LEN					130
#define DL_RECNO_SUBDIR_START_OFFSET	(DL_DIRNAME_OFFSET + DL_DIRNAME_LEN)			// サブディレクトリの先頭レコード番号
#define DL_RECNO_SUBDIR_START_LEN		2
#define DL_NUM_SUBDIR_OFFSET			(DL_RECNO_SUBDIR_START_OFFSET + DL_RECNO_SUBDIR_START_LEN)		// サブディレクトリ数
#define DL_NUM_SUBDIR_LEN				1
#define DL_RECNO_PARENTDIR_OFFSET		(DL_NUM_SUBDIR_OFFSET + DL_NUM_SUBDIR_LEN)		// 親ディレクトリのレコード番号
#define DL_RECNO_PARENTDIR_LEN			2
#define DL_RECNO_GRANDPARENTDIR_OFFSET	(DL_RECNO_PARENTDIR_OFFSET + DL_RECNO_PARENTDIR_LEN)	// 親々ディレクトリのレコード番号
#define DL_RECNO_GRANDPARENTDIR_LEN		2
#define DL_NUM_SNDFILE_OFFSET			(DL_RECNO_GRANDPARENTDIR_OFFSET + DL_RECNO_GRANDPARENTDIR_LEN)		// MP3ファイル数
#define DL_NUM_SNDFILE_LEN				1
#define DL_RECNO_SNDFILE_OFFSET			(DL_NUM_SNDFILE_OFFSET + DL_NUM_SNDFILE_LEN)	// MP3ファイルリスト先頭レコード番号
#define DL_RECNO_SNDFILE_LEN			2
#define DL_FILLER_LEN					80
#define DL_IMAGEDATA_OFFSET				RECLEN_DIRLIST		// イメージデータ
#define DL_IMAGEDATA_LEN				512
#define DL_IMAGEDATAHALF_LEN			256

// 音源ファイルリスト
#define	SNDFILELISTFILE					"/SNDLST18.DAT"		// MP3ファイルリスト(Version=16)
#define	RECLEN_MP3LIST					352UL					// イメージ部分を除くレコード長
#define	RECLEN_MP3LIST_FULL				1120UL					// レコード長
#define ML_DISPLAYNAME_OFFSET			0
#define ML_DISPLAYNAME_LEN				48
#define ML_FILENAME_OFFSET				(ML_DISPLAYNAME_OFFSET + ML_DISPLAYNAME_LEN)
#define ML_FILENAME_LEN					258
#define ML_FILLER_LEN					46
#define ML_IMAGEDATA_OFFSET				RECLEN_MP3LIST
#define ML_IMAGEDATA_LEN				768UL
#define ML_IMAGEDATAHALF_LEN			384UL

// ランダム再生順ファイル
#define	RANDOMORDERFILE					"/RANDOM18.DAT"
#define	RECLEN_RANDOMLIST				4
#define	RD_UNIT_OFFSET					0
#define	RD_UNIT_LEN						1
#define	RD_DIRRECNO_OFFSET				(RD_UNIT_OFFSET + RD_UNIT_LEN)
#define	RD_DIRRECNO_LEN					2
#define	RD_TRACKNO_OFFSET				(RD_DIRRECNO_OFFSET + RD_DIRRECNO_LEN)
#define	RD_TRACKNO_LEN					1

// 履歴ファイル HISTFILE parameters
#define HISTFILE                		"/HIST18.DAT"
#define RECLEN_HISTFILE         		9UL
#define HS_LASTDIR_OFFSET       		0
#define HS_LASTDIR_LEN	       			2
#define HS_LASTTRACK_OFFSET				(HS_LASTDIR_OFFSET + HS_LASTDIR_LEN)
#define HS_LASTTRACK_LEN				1
#define HS_LASTBYTE_OFFSET				(HS_LASTTRACK_OFFSET + HS_LASTTRACK_LEN)
#define HS_LASTBYTE_LEN             	4
#define	HS_RANDOMREC_OFFSET				(HS_LASTBYTE_OFFSET + HS_LASTBYTE_LEN)
#define	HS_RANDOMREC_LEN				2

// ボリュームファイル VOLUMEFILE parameters
#define VOLUMEFILE              		"VOLUME18.DAT"		// ディレクトリごとに作成
#define	DEFAULT_GAIN					-20

// メッセージファイル(MKFLMP3のCreateMessageStringFile()参照)
#define MSGSTRFILE						"/MSGSTR18.DAT"
#define MS_IMAGEDATA_LEN				256UL
#define MS_IMAGEDATAHALF_LEN			128UL
#define	MSGNO_MP3_18					0
#define	MSGNO_DIRLST_ERR				1
#define	MSGNO_SNDLST_ERR				2
#define	MSGNO_RANDOM_ERR				3
#define	MSGNO_MP3_ERR					4
#define	MSGNO_NO_HISTFILE				5

// I2C
#define SSD1306_I2C_CLK     	100
#define PIN_I2C_SDA				26
#define PIN_I2C_SCL				27
#define SSD1306_I2C_ADDR    	_u(0x3C)

// OLED
#define SSD1306_HEIGHT      	64
#define SSD1306_WIDTH       	128
#define LCD_STRING_LINES		4		// LCDに表示する文字列行数(固定値)
#define MAX_SELDIR_ITEM			4		// DIR選択画面で一度に表示されるDIR数の最大数
#define MAX_SELMP3_ITEM         4       // 曲選択画面で一度に表示される数

// SSD1306 commands (see datasheet)
#define SSD1306_SET_MEM_MODE        _u(0x20)
#define SSD1306_SET_COL_ADDR        _u(0x21)
#define SSD1306_SET_PAGE_ADDR       _u(0x22)
#define SSD1306_SET_HORIZ_SCROLL    _u(0x26)
#define SSD1306_SET_SCROLL          _u(0x2E)

#define SSD1306_SET_DISP_START_LINE _u(0x40)

#define SSD1306_SET_CONTRAST        _u(0x81)
#define SSD1306_SET_CHARGE_PUMP     _u(0x8D)

#define SSD1306_SET_SEG_REMAP       _u(0xA0)
#define SSD1306_SET_ENTIRE_ON       _u(0xA4)
#define SSD1306_SET_ALL_ON          _u(0xA5)
#define SSD1306_SET_NORM_DISP       _u(0xA6)
#define SSD1306_SET_INV_DISP        _u(0xA7)
#define SSD1306_SET_MUX_RATIO       _u(0xA8)
#define SSD1306_SET_DISP            _u(0xAE)
#define SSD1306_SET_PAGESTARTADDRESS _u(0xB0)
#define SSD1306_SET_COM_OUT_DIR     _u(0xC0)
#define SSD1306_SET_COM_OUT_DIR_FLIP _u(0xC0)

#define SSD1306_SET_DISP_OFFSET     _u(0xD3)
#define SSD1306_SET_DISP_CLK_DIV    _u(0xD5)
#define SSD1306_SET_PRECHARGE       _u(0xD9)
#define SSD1306_SET_COM_PIN_CFG     _u(0xDA)
#define SSD1306_SET_VCOM_DESEL      _u(0xDB)

#define SSD1306_PAGE_HEIGHT         _u(8)
#define SSD1306_NUM_PAGES           (SSD1306_HEIGHT / SSD1306_PAGE_HEIGHT)
#define SSD1306_BUF_LEN             (SSD1306_NUM_PAGES * SSD1306_WIDTH)

#define SSD1306_WRITE_MODE         _u(0xFE)
#define SSD1306_READ_MODE          _u(0xFF)

// ファイル読み込み用バッファサイズ
#if RECLEN_DIRLIST_FULL > RECLEN_MP3LIST_FULL
	#define	REC_BUF_SIZE	RECLEN_DIRLIST_FULL
#else
	#define	REC_BUF_SIZE	RECLEN_MP3LIST_FULL
#endif

// イベント/制御
static FATFS 		fs;
volatile int 		EventSw;		// イベントが発生したSWのGP番号
static volatile unsigned int	CHATTERING_TMR;
static volatile unsigned int	ScrollTimer;
volatile uint32_t	AutoPoffTimer;
static struct repeating_timer RpTimer;
static uint8_t		NowPlaying;
uint8_t				RandomPlaying;							// 1=ランダム再生中
static uint8_t      RecHistory[RECLEN_HISTFILE];			// 履歴ファイルレコード
static uint8_t		*PageImageBuff[SSD1306_NUM_PAGES];		// OLEDのPAGEごとのイメージデータ
static uint8_t		Page0Image[DL_IMAGEDATAHALF_LEN + 1];	// １行目はディレクトリ名をスクロール表示することを想定
static uint8_t		Page1Image[DL_IMAGEDATAHALF_LEN + 1];
static uint8_t		Page2Image[SSD1306_WIDTH + 1];			// ２行目以降は折り返し表示
static uint8_t		Page3Image[SSD1306_WIDTH + 1];
static uint8_t		Page4Image[SSD1306_WIDTH + 1];
static uint8_t		Page5Image[SSD1306_WIDTH + 1];
static uint8_t		Page6Image[SSD1306_WIDTH + 1];
static uint8_t		Page7Image[SSD1306_WIDTH + 1];
static int			PageImageWidth[SSD1306_NUM_PAGES];		// 上記PageXImageバッファのイメージデータ部分サイズ
static int			PageImageLength[SSD1306_NUM_PAGES];		// 実際のイメージの長さ
static int			MaxScrollX;		// 最大スクロール量
static int			ScrollX;

typedef enum {PD_SEQUENTIAL=0, PD_ONLY_ONE } playdirectory_t;		// ディレクトリ下の音源を順次再生するか、特定のファイルだけを再生するか

// この値によってSDKの機能が変わってくるので確認のため
#if PICO_ON_DEVICE == 1
	#pragma message("PICO_ON_DEVICE is 1")
#elif PICO_ON_DEVICE == 0
	#pragma message("PICO_ON_DEVICE is 0")
#else
	#pragma message("PICO_ON_DEVICE is ?")
#endif

// 関数宣言
void InitApp();
int SelectDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3);
int SelectSoundfile(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno);
int PlayOneDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno, uint32_t start_byte, playdirectory_t mode);
int PlayRandom(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3);
int PlayOneMp3(const TCHAR *mp3file, uint16_t dirno, uint8_t trackno, uint32_t start_byte, double *gain);
int PauseAndSelectItem();
int ReadHistory(FIL *Fil);
int SaveHistory(FIL *Fil);
double GetLocalGain(TCHAR *volumefile, FIL *Fil);
void SaveLocalGain(TCHAR *volumefile, FIL *Fil, double gain);
WCHAR *MakeUnicodeString(const char *s);
void SSD1306_init();
void SSD1306_send_cmd_list(uint8_t *buf, int num);
void SSD1306_send_cmd(uint8_t cmd);
void DispLcdFullImage();
void DispMsg(int line, int msgno, int inv, FIL *Fil);
void SetMenuItemSelected(uint8_t line);
void InitImageBuff();
void ClearImageBuff();
void ClearPageImageBuff(int page);
void SetPageImageBuff(int page, uint8_t *buf, int len);
void DispScrollLine(int line);
void error_blink(uint led, int count);


int main()
{
    uint8_t     RecBuf[REC_BUF_SIZE];
    TCHAR		testfile[32];
    FIL 		FilDL;
    FIL 		FilML;
    FIL 		Fil3;
    FRESULT		res;     /* FatFs return code */
	int			ret;
	
	InitApp();
   	DispMsg(0, MSGNO_MP3_18, 0, &Fil3);
   	
    // DIR構成ファイル
    res = f_open(&FilDL, MakeUnicodeString(DIRSTRUCTUREFILE), FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
    	printf("f_open error\n\r");
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, &Fil3);
    	sleep_ms(3000);
        goto POFF;
    } else {
    	printf("DIRSTRUCTUREFILE open ok\n");
    }
    // MP3ファイルリスト
    res = f_open(&FilML, MakeUnicodeString(SNDFILELISTFILE), FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
    	printf("f_open error\n\r");
    	DispMsg(0, MSGNO_SNDLST_ERR, 0, &Fil3);
    	sleep_ms(3000);
        goto POFF;
    } else {
    	printf("SNDFILELISTFILE open ok\n");
    }
    // 履歴ファイル
    ReadHistory(&Fil3);
    
    do {
		ret = SelectDirectory(RecBuf, &FilDL, &FilML, &Fil3);
		if (ret == START_RANDOM){
			ret = PlayRandom(RecBuf, &FilDL, &FilML, &Fil3);
			if (ret == POWEROFF){
				break;
			}
		} else break;
	} while (1);
	
    f_close(&FilDL);
    f_close(&FilML);

    // OK blink
    POFF:
    printf("POFF\n");
    gpio_put(PIN_POWER_ON, 0);
	while (1){
        error_blink(PIN_LED_ONBOARD, 1);
	}
    return 0;
}


void gpio_callback(uint gpio, uint32_t events) {
	if (CHATTERING_TMR == 0){
	    CHATTERING_TMR = 1;
	    printf("GPIO %d \n", gpio);
	    EventSw = gpio;
	}
}


bool repeating_timer_callback(struct repeating_timer *t) {
	static int STOP_COUNT = 10;		// 横スクロールの停止時間
	
	AutoPoffTimer++;
	if (CHATTERING_TMR > 0){
		CHATTERING_TMR++;
		if (CHATTERING_TMR == 4) CHATTERING_TMR = 0;
	}
	if (NowPlaying == 1 && MaxScrollX > 0){
		ScrollTimer++;
		if (ScrollTimer == STOP_COUNT){		// 停止時間すぎた
			ScrollTimer = 0;
			ScrollX += 8;			// 8ピクセル進める
			STOP_COUNT = 10;		// 1秒
			if (ScrollX > MaxScrollX){
				ScrollX = 0;
				STOP_COUNT = 30;	// 3秒
			}
			DispScrollLine(0);		// 1行目＝ディレクトリ名を横スクロール
		}
	}
    return true;
}


/*
 *	各種初期化(IO/SD/I2S/I2C)
 */
void InitApp(){
    FRESULT res;     /* FatFs return code */
    
    stdio_init_all();

	EventSw = -1;
	CHATTERING_TMR = 0;
	AutoPoffTimer = 0;
    NowPlaying = 0;
	
    // Initialise UART 0
    uart_init(uart0, 115200);
    // Set the GPIO pin mux to the UART - 0 is TX, 1 is RX
    gpio_set_function(0, GPIO_FUNC_UART);
    gpio_set_function(1, GPIO_FUNC_UART);

    gpio_init(PIN_POWER_ON);
    gpio_set_dir(PIN_POWER_ON, GPIO_OUT);	// pico-sdk\src\host\hardware_gpio\include\hardware\gpio.h
    gpio_put(PIN_POWER_ON, 1);
	
	// I2C
    i2c_init(i2c1, SSD1306_I2C_CLK * 1000);
    gpio_set_function(PIN_I2C_SDA, GPIO_FUNC_I2C);
    gpio_set_function(PIN_I2C_SCL, GPIO_FUNC_I2C);
    gpio_pull_up(PIN_I2C_SDA);
    gpio_pull_up(PIN_I2C_SCL);
	sleep_ms(100);		// ここに待ちを入れないとI2Cが安定してないようだ
    // OLED
    InitImageBuff();
	SSD1306_init();
	ClearImageBuff();
    DispLcdFullImage();
    
    add_repeating_timer_ms(TIMER_INTERVAL_MS, repeating_timer_callback, NULL, &RpTimer);
    
    gpio_init(PIN_SW_POWER_OFF);
    gpio_set_dir(PIN_SW_POWER_OFF, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_POWER_OFF, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
    
    gpio_init(PIN_SW_DOWN);
    gpio_set_dir(PIN_SW_DOWN, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_DOWN, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_UP);
    gpio_set_dir(PIN_SW_UP, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_UP, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_SELECT);
    gpio_set_dir(PIN_SW_SELECT, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_SELECT, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_RETURN);
    gpio_set_dir(PIN_SW_RETURN, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_RETURN, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_PLAY);
    gpio_set_dir(PIN_SW_PLAY, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_PLAY, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_RANDOM);
    gpio_set_dir(PIN_SW_RANDOM, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_RANDOM, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_PGDOWN);
    gpio_set_dir(PIN_SW_PGDOWN, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_PGDOWN, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_PGUP);
    gpio_set_dir(PIN_SW_PGUP, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_PGUP, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_SW_OPTION);
    gpio_set_dir(PIN_SW_OPTION, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_SW_OPTION, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_RE_DOWN);
    gpio_set_dir(PIN_RE_DOWN, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_RE_DOWN, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_RE_UP);
    gpio_set_dir(PIN_RE_UP, GPIO_IN);
    gpio_set_irq_enabled_with_callback(PIN_RE_UP, GPIO_IRQ_EDGE_FALL, true, &gpio_callback);
	
    gpio_init(PIN_LED_ONBOARD);
    gpio_set_dir(PIN_LED_ONBOARD, GPIO_OUT);

	printf("start\n");
	
    res = f_mount(&fs, MakeUnicodeString(""), 1);
    if (res != FR_OK) {
        printf("mount error %d\n", res);
        error_blink(PIN_LED_ONBOARD, 1);
    }
    printf("mount ok\n");
    
    
    init_audio();		// audio_i2s.c
    //ap_producer = init_audio(44100, 2, 4);		// producer
	//init_i2s(44100, 2);	
}


/*
 * ディレクトリの選択操作
 * 	レコード番号：ヘッダレコードを０とする。以降、ルートディレクトリを１、など。
 * 	RecBufはDIR構成ファイルまたはMP3ファイルリストの１レコードを取得できるサイズを持つ。
 * 	FilDL:DIR構成ファイル、FilML:MP3ファイルリスト、いずれもf_open済み。Fil3はその他任意のファイルでそのつどf_open/fcloseする。
 */
int SelectDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3){
    FRESULT     res;
    uint32_t    ofsSubDirStart;
    uint16_t    rowSubDirStart;			// 現在対象のディレクトリの先頭サブディレクトリのレコード番号
    uint16_t    numSubDir;				// サブディレクトリ数
    uint16_t    idxSubDir;				// 選択ページ先頭に表示中のDIR (0,1,..., numSubDir-1)
    uint8_t     idxCurSubDirItem;   	// 現在表示ページで選択中のDIR(0,1,2,3)
    uint8_t     idxCurSubDirItemBk;		// 上記の一時的退避
    uint8_t		NumSelItem;				// １ページに表示する最大項目数
    int        	ret;
    uint8_t     idxDirDepth;
    uint16_t    stackIdxSubDir[8];
    uint8_t     stackIdxCurrentSubDirItem[8];
    uint16_t    dirno[MAX_SELDIR_ITEM];
    uint16_t    numSubDirTab[MAX_SELDIR_ITEM];
    uint16_t    numSndFile[MAX_SELDIR_ITEM];
    uint16_t    recNoFirstSubdir[MAX_SELDIR_ITEM];
    uint8_t     readdlflg;      // DIR LIST ファイルを読む必要があるか？
    uint8_t		i;
    uint8_t		items = 0;
    UINT    	br;

//    printf("SelectDirectory start\r\n");
    ret = FR_OK;

    // Skip HEADER info
    res = f_lseek(FilDL, RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	sleep_ms(3000);
    	printf("f_lseek error\n\r");
        return res;
    }

    // Read ROOT info
    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST, &br);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	sleep_ms(3000);
    	printf("f_read error (root)\r\n");
        return res;
    }

    // ROOTディレクトリの最初のサブディレクトリのレコード番号
	memcpy(&rowSubDirStart, RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);
    ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
    numSubDir = *(RecBuf + DL_NUM_SUBDIR_OFFSET);
    idxSubDir = 0;
    idxCurSubDirItem = 0;
    idxCurSubDirItemBk = -1;
    idxDirDepth = 0;
    readdlflg = TRUE;
    NumSelItem = MAX_SELDIR_ITEM;

    // end point of ROOT info
    res = f_lseek(FilDL, 2*RECLEN_DIRLIST_FULL);		// ヘッダレコードとルートディレクトリレコードを読み飛ばす
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	sleep_ms(3000);
    	printf("f_lseek error (end of root)\r\n");
        return res;
    }

    while (1){
    	if (readdlflg == TRUE){
    		items = 0;
			res = f_lseek(FilDL, ofsSubDirStart + (uint32_t)idxSubDir * RECLEN_DIRLIST_FULL);
			for (i=0; i<NumSelItem; i++){
				if ((idxSubDir + i) < numSubDir){
					if ((res = f_read(FilDL, RecBuf, RECLEN_DIRLIST_FULL, &br)) == FR_OK){
						printf("SelectDirectory %d\n", i);
						dirno[i] = rowSubDirStart + idxSubDir + i;
                        numSubDirTab[i] = *(RecBuf + DL_NUM_SUBDIR_OFFSET);
						numSndFile[i] = *(RecBuf + DL_NUM_SNDFILE_OFFSET);
						memcpy(&recNoFirstSubdir[i], RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);
						items++;
						// イメージ
						SetPageImageBuff(2*i, RecBuf + RECLEN_DIRLIST, SSD1306_WIDTH);
						SetPageImageBuff(2*i + 1, RecBuf + RECLEN_DIRLIST + DL_IMAGEDATAHALF_LEN, SSD1306_WIDTH);
					} else {
						// エラー
				    	DispMsg(i, MSGNO_DIRLST_ERR, 0, Fil3);
					}
				} else {
					dirno[i] = 0;
					numSndFile[i] = 0;
					recNoFirstSubdir[i] = 0;
					ClearPageImageBuff(2*i);
					ClearPageImageBuff(2*i + 1);
				}
			}
			SetMenuItemSelected(idxCurSubDirItem);
        	DispLcdFullImage();
    	}

		ret = PauseAndSelectItem();
	    readdlflg = TRUE;
		if (ret == SWUP_PRESSED){
			// 同一ページ内の上の行へ、または、前ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == 0){
				if (idxSubDir >= NumSelItem){
					idxSubDir -= NumSelItem;	// 前ページへ
					idxCurSubDirItem = NumSelItem - 1;
				} else {
					readdlflg = FALSE;
				}
			} else {
				idxCurSubDirItem--;
				readdlflg = FALSE;
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk);
				SetMenuItemSelected(idxCurSubDirItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWDOWN_PRESSED){
			// 同一ページ内の下の行へ、または、次ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == (LCD_STRING_LINES - 1)){
				if ((idxSubDir + NumSelItem) < numSubDir){
					idxSubDir += NumSelItem;	// 次のページへ
					idxCurSubDirItem = 0;
				} else {
					readdlflg = FALSE;
				}
			} else {
				if (idxCurSubDirItem < (items - 1)){
					idxCurSubDirItem++;
					readdlflg = FALSE;
				} else {
					readdlflg = FALSE;
				}
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk);
				SetMenuItemSelected(idxCurSubDirItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWPAGEUP_PRESSED || ret == REUP_PRESSED){		// 改ページ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if (idxSubDir >= NumSelItem){
				idxSubDir -= NumSelItem;	// 前ページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPAGEDOWN_PRESSED || ret == REDOWN_PRESSED){	// 改ページ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxSubDir + NumSelItem) < numSubDir){
				idxSubDir += NumSelItem;	// 次のページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPLAY_PRESSED){		// PLAY
			if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, dirno[idxCurSubDirItem], 0, 0, PD_SEQUENTIAL);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					//break;
				}
			} else if (idxDirDepth == 0) {	// ROOTディレクトリでは前回再生の再開（ROOTディレクトリにはMP3ファイルを置かない）
				uint16_t	last_dirno;
				uint8_t		last_trackno;
				uint32_t	last_bytes;

				memcpy(&last_dirno, &RecHistory[HS_LASTDIR_OFFSET], HS_LASTDIR_LEN);
				last_trackno = RecHistory[HS_LASTTRACK_OFFSET];
				memcpy(&last_bytes, &RecHistory[HS_LASTBYTE_OFFSET], HS_LASTBYTE_LEN);
				if (last_dirno > 1){	// 前回再生の履歴が存在すること
					ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, last_dirno, last_trackno, last_bytes, PD_SEQUENTIAL);
					if (ret == POWEROFF){
						break;
					} else if (ret == GOTO_MAINMENU){
						//break;
					}
				}
			} else {
				//
		    	printf("No mp3 files\n\r");
			}
		} else if (ret == SWRANDOM_PRESSED){	// RANDOM
			ret = START_RANDOM;
			break;
		} else if (ret == SWSELECT_PRESSED){  // 下位DIRへ (SKIP forward directory)
			if (recNoFirstSubdir[idxCurSubDirItem] > 1){		// 下位DIRが存在する
				rowSubDirStart = recNoFirstSubdir[idxCurSubDirItem];
				ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
				numSubDir = numSubDirTab[idxCurSubDirItem];		// *(RecDirList + DL_NUM_SUBDIR);
				// push current idxSubDir to stack
				stackIdxSubDir[idxDirDepth] = idxSubDir;
				stackIdxCurrentSubDirItem[idxDirDepth] = idxCurSubDirItem;
				idxDirDepth++;
				idxSubDir = 0;
				idxCurSubDirItem = 0;
		        //NumSelItem = MAX_SELDIR_ITEM;		//初期値に戻さない
			} else if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				res = f_lseek(FilDL, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL);
			    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST, &br);

				ret = SelectSoundfile(RecBuf, FilDL, FilML, Fil3, dirno[idxCurSubDirItem], 0);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					//break;
				}
			}
		} else if (ret == SWRETURN_PRESSED){  // 上位DIRへ (GOTO Mainmenu)
			uint16_t    r;
			uint16_t    r2;

			//r = *((uint16_t *)(RecDirList + DL_RECNO_GRANDPARENTDIR));   // rec no of grand parent
			//r = P8U16(RecDirList, DL_RECNO_GRANDPARENTDIR);		// rec no of grand parent
			// RecDirListは表示中のディレクトリのどれかのDIR構成ファイルレコードを保持していることが前提！
			memcpy(&r,  RecBuf + DL_RECNO_GRANDPARENTDIR_OFFSET, DL_RECNO_GRANDPARENTDIR_LEN);	// rec no of grand parent
			memcpy(&r2, RecBuf + DL_RECNO_PARENTDIR_OFFSET, DL_RECNO_PARENTDIR_LEN);
			// 親々ディレクトリのレコード番号と親ディレクトリのレコード番号が一致しているときは、トップレベルのDIR階層
			if (r == r2){
				//なにもしない
				//ret = GOTO_MAINMENU;
				//break;
			} else {
				// Go to grand parent record
				res = f_lseek(FilDL, (uint32_t)r * RECLEN_DIRLIST_FULL);
				if (res != FR_OK) return res;
				// Read grand parent record
				//res = f_read_DIRLISTFILE(Fil, RecDirList);
			    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST, &br);
				if (res != FR_OK) return res;
				// the first "uncle叔父" record
				//r = *((uint16_t *)(RecDirList + DL_RECNO_SUBDIR_START));
				//r = P8U16(RecDirList, DL_RECNO_SUBDIR_START);
				memcpy(&r, RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);
				if (r > 1){
					rowSubDirStart = r;
					ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
					numSubDir = *(RecBuf + DL_NUM_SUBDIR_OFFSET);
					// pop idxSubDir from stack -> This is parent
					idxDirDepth--;
					idxSubDir = stackIdxSubDir[idxDirDepth];
					idxCurSubDirItem = stackIdxCurrentSubDirItem[idxDirDepth];
					if (idxSubDir >= numSubDir) idxSubDir = 0;
					readdlflg = TRUE;
				} else {
					//なにもしない
					//ret = GOTO_MAINMENU;
					//break;
				}
			}
			//readdlflg = TRUE;
		} else if (ret == SWOPTION_PRESSED){      //
			// オプション機能案
			// コントラスト設定
			readdlflg = TRUE;
		} else if (ret == SWPOFF_PRESSED){
			ret = POWEROFF;
			break;
		}

    }

    printf("SelectDirectory end\r\n");
    return ret;
}


/*
 * 音源ファイルの選択操作
 * 	FilDL:DIR構成ファイル、FilML:MP3ファイルリスト、いずれもf_open済み。Fil3はその他任意のファイルでそのつどf_open/fcloseする。
 */
int SelectSoundfile(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno){
    FRESULT     res;
    int		    ret;
    //uint16_t    rowSubDirStart;	// 対象ディレクトリの先頭サブディレクトリのレコード番号
    uint16_t 	numsndfile;		// 対象ディレクトリのMP3ファイル数
    uint16_t	rowSndStart;	// 対象ディレクトリのMP3ファイルリスト先頭レコード番号
    uint8_t     trackno;		// 選択中の曲番 (0,1,2,...,numsndfile-1)
    uint8_t     idxtrack;       // 選択操作中のページ先頭の曲番 (0,1,2,...,numsndfile-1)
    uint8_t		NumSelItem;		// １ページに表示する項目数
    UINT        br;
    uint16_t    dirbytes;
    uint8_t     readdlflg;      // MP3 LIST ファイルを読む必要があるか？
    uint8_t		i;
    uint8_t		items = 0;
    uint16_t    idxCurTitleItem;
    uint16_t    idxCurTitleItemBk;

    ret = FR_OK;

	printf("SelectSoundfile start\n\r");
	res = f_lseek(FilDL, (uint32_t)(dirno) * RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	sleep_ms(3000);
    	printf("f_lseek error\n\r");
        return res;
    }
    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST_FULL, &br);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	sleep_ms(3000);
    	printf("f_read error\n\r");
        return res;
    }
    memcpy(&dirbytes, RecBuf + DL_FULLDIRBYTES_OFFSET, DL_FULLDIRBYTES_LEN);
    numsndfile = RecBuf[DL_NUM_SNDFILE_OFFSET];
    if (numsndfile == 0){
    	printf("no mp3 files\n\r");
    	return FR_OK;
    }
	memcpy(&rowSndStart, RecBuf + DL_RECNO_SNDFILE_OFFSET, DL_RECNO_SNDFILE_LEN);
	//memcpy(&rowSubDirStart, RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);

    if (start_trackno >= numsndfile) start_trackno = 0;
    trackno = start_trackno;
    idxtrack = 0;
    idxCurTitleItem = 0;
    readdlflg = TRUE;
    NumSelItem = MAX_SELMP3_ITEM;

    while (1){
		if (readdlflg == TRUE){		//
			// 選択肢を表示
			items = 0;
			ClearImageBuff();
			res = f_lseek(FilML, (uint32_t)(rowSndStart + idxtrack) * RECLEN_MP3LIST_FULL);
			for (i=0; i<NumSelItem; i++){
				if ((idxtrack + i) < numsndfile){
					if ((res = f_read(FilML, RecBuf, RECLEN_MP3LIST_FULL, &br)) == FR_OK){
						items++;
						// イメージ
						SetPageImageBuff(2*i, RecBuf + RECLEN_MP3LIST, SSD1306_WIDTH);
						SetPageImageBuff(2*i + 1, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN, SSD1306_WIDTH);
					} else {
						// エラー
						DispMsg(2*i, MSGNO_SNDLST_ERR, 0, Fil3);
					}
				}
			}
			SetMenuItemSelected(idxCurTitleItem);
			DispLcdFullImage();
		}

		ret = PauseAndSelectItem();
		readdlflg = TRUE;
		if (ret == SWUP_PRESSED){
			idxCurTitleItemBk = idxCurTitleItem;
			if (idxCurTitleItem == 0){
				//idxCurTitleItem = NumSelItem - 1;
				if (idxtrack >= NumSelItem){
					idxtrack -= NumSelItem;
					//idxCurTitleItem = 0;
					idxCurTitleItem = NumSelItem - 1;
				} else {
					readdlflg = FALSE;
				}
			} else {
				idxCurTitleItem--;
				readdlflg = FALSE;
			}
			if (readdlflg == FALSE && idxCurTitleItem != idxCurTitleItemBk){
				SetMenuItemSelected(idxCurTitleItemBk);
				SetMenuItemSelected(idxCurTitleItem);
				DispLcdFullImage();
			}
		} else if (ret == SWDOWN_PRESSED){
			idxCurTitleItemBk = idxCurTitleItem;
			if (idxCurTitleItem < (NumSelItem - 1)){
				if ((idxtrack + idxCurTitleItem) < (numsndfile - 1)){
					idxCurTitleItem++;
					readdlflg = FALSE;
				} else {
					readdlflg = FALSE;
				}
			} else {
				//idxCurTitleItem = 0;
				if ((idxtrack + NumSelItem) < numsndfile){
					idxtrack += NumSelItem;
					idxCurTitleItem = 0;
				} else {
					readdlflg = FALSE;
				}
			}
			if (readdlflg == FALSE && idxCurTitleItem != idxCurTitleItemBk){
				SetMenuItemSelected(idxCurTitleItemBk);
				SetMenuItemSelected(idxCurTitleItem);
				DispLcdFullImage();
			}
		} else if (ret == SWPAGEUP_PRESSED || ret == REUP_PRESSED){		// 改ページ
			if (idxtrack >= NumSelItem){
				idxtrack -= NumSelItem;
				idxCurTitleItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPAGEDOWN_PRESSED || ret == REDOWN_PRESSED){	// 改ページ
			if ((idxtrack + NumSelItem) < numsndfile){
				idxtrack += NumSelItem;
				idxCurTitleItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPLAY_PRESSED){
			trackno = idxtrack + idxCurTitleItem;
			ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, dirno, trackno, 0, PD_SEQUENTIAL);
			if (ret == POWEROFF){
				break;
			} else if (ret == GOTO_MAINMENU){
				break;
			}
		} else if (ret == SWRETURN_PRESSED){
			ret = PLAY_OTHER_DIR;
			break;			// 現ディレクトリを抜ける。
		} else if (ret == SWOPTION_PRESSED){
			readdlflg = TRUE;
		} else if (ret == SWPOFF_PRESSED){
			ret = POWEROFF;
			break;
		}

    }

	printf("SelectSoundfile end\n\r");
    return ret;
}


/*
 * ディレクトリ下のMP3ファイルを順次再生
 * 	FilDL:DIR構成ファイル、FilML:MP3ファイルリスト、いずれもf_open済み。Fil3はその他任意のファイルでそのつどf_open/fcloseする。
 */
int PlayOneDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno, uint32_t start_byte, playdirectory_t mode){
    FRESULT     res;
    int		    ret;
    uint8_t		filefullpath[DL_FULLDIR_LEN + ML_FILENAME_LEN + 2];
    uint16_t 	numsndfile;		// 対象ディレクトリのMP3ファイル数
    uint16_t	rowSndStart;	// 対象ディレクトリのMP3ファイルリスト先頭レコード番号
    uint8_t     trackno;		// 再生中の曲番 (0,1,2,...,numsndfile-1)
    UINT        br;
    uint16_t    dirbytes;
    uint8_t     execplayflg = 0;
    int			vflen;
    double		gain;

    ret = FR_OK;

	printf("PlayOneDirectory start\n\r");
	res = f_lseek(FilDL, (uint32_t)(dirno) * RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	sleep_ms(3000);
    	printf("f_lseek error\n\r");
        return res;
    }
    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST_FULL, &br);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	sleep_ms(3000);
    	printf("f_read error\n\r");
        return res;
    }
    memcpy(filefullpath, RecBuf + DL_FULLDIR_OFFSET, DL_FULLDIR_LEN);
    memcpy(&dirbytes, RecBuf + DL_FULLDIRBYTES_OFFSET, DL_FULLDIRBYTES_LEN);
    numsndfile = RecBuf[DL_NUM_SNDFILE_OFFSET];
    if (numsndfile == 0){
    	printf("no mp3 files\n\r");
    	return FR_OK;
    }
    printf("numsndfile=%d dirbytes=%d\n", numsndfile, dirbytes);
	memcpy(&rowSndStart, RecBuf + DL_RECNO_SNDFILE_OFFSET, DL_RECNO_SNDFILE_LEN);
	// １行目にディレクトリ名
	SetPageImageBuff(0, RecBuf + RECLEN_DIRLIST, DL_IMAGEDATAHALF_LEN);
	SetPageImageBuff(1, RecBuf + RECLEN_DIRLIST + DL_IMAGEDATAHALF_LEN, DL_IMAGEDATAHALF_LEN);
	// １行目の横スクロール条件
	if (PageImageLength[0] > SSD1306_WIDTH || PageImageLength[1] > SSD1306_WIDTH){
		int len = (PageImageLength[0] > PageImageLength[1] ? PageImageLength[0] : PageImageLength[1]);
		MaxScrollX = len - SSD1306_WIDTH;
		ScrollX = 0;
	} else {
		MaxScrollX = 0;
	}
    // ディレクトリパスとファイル名との間の区切り (UNICODEの'\')
    filefullpath[dirbytes++] = '\\';
    filefullpath[dirbytes++] = 0;
    if (start_trackno >= numsndfile) start_trackno = 0;
	
    // ボリュームファイルによる音量設定
    vflen = 2*strlen(VOLUMEFILE);		// UNICODE長
    memcpy(filefullpath + dirbytes, MakeUnicodeString(VOLUMEFILE), vflen);
    memset(filefullpath + dirbytes + vflen, 0, 2);
    gain = GetLocalGain((TCHAR *)filefullpath, Fil3);

    for (trackno=start_trackno; trackno<numsndfile; trackno++){
    	printf("trackno=%d\n", trackno);
    	if (mode == PD_ONLY_ONE && trackno != start_trackno) continue;
        if (f_lseek(FilML, (uint32_t)(rowSndStart + trackno)*(RECLEN_MP3LIST_FULL)) != FR_OK) break;
        if (f_read(FilML, RecBuf, RECLEN_MP3LIST_FULL, &br) != FR_OK) break;
		// ２，３，４行目に曲名表示
		SetPageImageBuff(2, RecBuf + RECLEN_MP3LIST, SSD1306_WIDTH);
		SetPageImageBuff(3, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN, SSD1306_WIDTH);
		SetPageImageBuff(4, RecBuf + RECLEN_MP3LIST + SSD1306_WIDTH, SSD1306_WIDTH);
		SetPageImageBuff(5, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN + SSD1306_WIDTH, SSD1306_WIDTH);
		SetPageImageBuff(6, RecBuf + RECLEN_MP3LIST + 2*SSD1306_WIDTH, SSD1306_WIDTH);
		SetPageImageBuff(7, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN + 2*SSD1306_WIDTH, SSD1306_WIDTH);
    	DispLcdFullImage();

        // Mp3 file full path
        memcpy(filefullpath + dirbytes, RecBuf + ML_FILENAME_OFFSET, ML_FILENAME_LEN);
        // Play a MP3 file
        ret = PlayOneMp3((TCHAR *)filefullpath, dirno, trackno, start_byte, &gain);
        if (ret == PLAY_NEXT_MP3){				// SWPLAY
            // 再生続行
        } else if (ret == PLAY_OTHER_DIR){		// SWRANDOM（再生中にRANDOMボタンが使えるのはランダム再生中だけ）
        	break;
        } else if (ret == POWEROFF){			// SWPOWER
            break;
        } else if (ret == SELECT_OTHER_MP3){	// SWRETURN
        	ret = GOTO_MAINMENU;
        	break;
        } else {
            execplayflg = 1;
        }
        start_byte = 0;
    }

    // ボリュームファイル保存
    if (execplayflg == 1){
        memcpy(filefullpath + dirbytes, MakeUnicodeString(VOLUMEFILE), vflen);
        memset(filefullpath + dirbytes + vflen, 0, 2);
        SaveLocalGain((TCHAR *)filefullpath, Fil3, gain);
    }
    // 履歴保存
    SaveHistory(Fil3);

    return ret;
}


/*
 * ランダム再生
 */
int PlayRandom(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3){
    FRESULT     res;
    int		    ret;
    FIL 		FilRL;
    uint8_t		rec_random[RECLEN_RANDOMLIST];
    uint8_t		unit;
    uint16_t	recno_random;
    uint16_t	recno_dir;
    uint8_t		trackno;
    UINT		bytes;
	uint32_t	last_bytes;

    ret = FR_OK;
	RandomPlaying = 1;

	printf("PlayRandom start\n");
	// 前回再生の再開から
	memcpy(&recno_dir, &RecHistory[HS_LASTDIR_OFFSET], HS_LASTDIR_LEN);
	trackno = RecHistory[HS_LASTTRACK_OFFSET];
	memcpy(&last_bytes, &RecHistory[HS_LASTBYTE_OFFSET], HS_LASTBYTE_LEN);
	if (recno_dir > 1){
		ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, recno_dir, trackno, last_bytes, PD_SEQUENTIAL);	// 常にPD_SEQUENTIAL
        if (ret == POWEROFF || ret == GOTO_MAINMENU){
        	goto end;
        }
	}
	
	// 次のランダム順
	res = f_open(&FilRL, MakeUnicodeString(RANDOMORDERFILE), FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		DispMsg(0, MSGNO_RANDOM_ERR, 0, Fil3);
		sleep_ms(2000);
       	goto end;
	} else {
    	printf("RANDOMORDERFILE open ok\n");
	}
	memcpy(&recno_random, RecHistory + HS_RANDOMREC_OFFSET,  HS_RANDOMREC_LEN);
    res = f_lseek(&FilRL, (DWORD)recno_random*RECLEN_RANDOMLIST);
	if (res != FR_OK){
	    res = f_lseek(&FilRL, 0);
	}

	do {
		res = f_read(&FilRL, rec_random, RECLEN_RANDOMLIST, &bytes);
		if (res != FR_OK || bytes != RECLEN_RANDOMLIST){
		    res = f_lseek(&FilRL, 0);
			if (res == FR_OK){
				recno_random = 0;
				continue;
			} else {
				break;
			}
		}
		unit = rec_random[RD_UNIT_OFFSET];
		memcpy(&recno_dir, &rec_random[RD_DIRRECNO_OFFSET], RD_DIRRECNO_LEN);
		trackno = rec_random[RD_TRACKNO_OFFSET];
		printf("recno=%d trackno=%d unit=%d\n", recno_dir, trackno, unit);
		recno_random++;
		memcpy(RecHistory + HS_RANDOMREC_OFFSET, &recno_random, HS_RANDOMREC_LEN);

		if (unit == PD_ONLY_ONE){
			ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, recno_dir, trackno, 0, PD_ONLY_ONE);
		} else {
			ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, recno_dir, 0, 0, PD_SEQUENTIAL);
		}

        if (ret == POWEROFF || ret == GOTO_MAINMENU){
        	break;
        } else {
        	// ランダム再生続行。ランダム再生順ファイルの次のレコードへ。
        }
	} while (1);
	f_close(&FilRL);
	
	end:
	RandomPlaying = 0;

	return ret;
}


/*
 * MP3再生
 * 	mp3file:MP3ファイルフルパス、dirno:ルートディレクトリ=0から始まる行番号
 */
int PlayOneMp3(const TCHAR *mp3file, uint16_t dirno, uint8_t trackno, uint32_t start_byte, double *gain){
  	struct player player;
    FRESULT     res;
    int         ret;

    ret = FR_OK;
	printf("PlayOneMp3 start %d\n", dirno);
	
  	player_init(&player);
	player.input.path = mp3file;
    res = f_open(&player.input.fd, mp3file, FA_READ);
    if (res == FR_OK) {
		// 途中再開
		if (start_byte > 0){
			if (f_lseek(&player.input.fd, start_byte) != FR_OK){      // go to restart point
				f_lseek(&player.input.fd, 0);
				start_byte = 0;
			} else if (f_eof(&player.input.fd)){
				goto end;
			}
		}
	    memcpy(RecHistory + HS_LASTDIR_OFFSET, &dirno, HS_LASTDIR_LEN);
	    *(uint16_t *)(RecHistory + HS_LASTTRACK_OFFSET) = trackno;
	    *(uint32_t *)(RecHistory + HS_LASTBYTE_OFFSET) = start_byte;
		NowPlaying = 1;
		ret = player_run(&player, (TCHAR *)mp3file, *gain);			// madplay.c
		if (ret != -1){
			uint32_t	p = f_tell(&player.input.fd);
			memcpy(RecHistory + HS_LASTBYTE_OFFSET, &p, HS_LASTBYTE_LEN);
			*gain = player.output.attamp_db;
		}
		NowPlaying = 0;
		end:
		f_close(&player.input.fd);
		player_finish(&player);
    } else {
    	printf("f_open error %d\n", res);
    	ret = -1;
    }
    

	return ret;
}


// 選択画面における入力待ち
int PauseAndSelectItem(){
	int		ret;

	printf("PauseAndSelectItem start");
	ret = 0;
	EventSw = -1;
	AutoPoffTimer = 0;
	do {
		if (EventSw != -1){
			ret = EventSw + 1;
			EventSw = -1;
			return ret;
		}
		if (AutoPoffTimer >= AUTOPOFF_PAUSE){
			return SWPOFF_PRESSED;
		}
	} while (1);
}


/*
 * 履歴ファイルを読み込む
 */
int ReadHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    memset(RecHistory, 0, RECLEN_HISTFILE);
    res = f_open(Fil, MakeUnicodeString(HISTFILE), FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        f_read(Fil, RecHistory, RECLEN_HISTFILE, &br);
        f_close(Fil);
    } else {
    	printf("ReadHistory f_open error\r\n");
    }

    return TRUE;
}


/*
 * 履歴ファイルを保存
 */
int SaveHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    printf("SaveHistory start \r\n");
    res = f_open(Fil, MakeUnicodeString(HISTFILE), FA_READ | FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
    	//printf("SaveHistory f_open OK \r\n");
        res = f_write(Fil, RecHistory, RECLEN_HISTFILE, &br);
        if (res == FR_OK){
        	//printf("SaveHistory f_write OK \r\n");
        }
        f_close(Fil);
        return TRUE;
    } else {
    	printf("SaveHistory f_open error \r\n");
    }
    return FALSE;
}


/*
 * ディレクトリごとに保存されているGainを取得
 */
double GetLocalGain(TCHAR *volumefile, FIL *Fil){
    FRESULT     res;
	double 		d;
    UINT    	br;
	
    res = f_open(Fil, volumefile, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        res = f_read(Fil, &d, sizeof(double), &br);
        if (res != FR_OK || br != sizeof(double)){
        	d = DEFAULT_GAIN;
        }
    	f_close(Fil);
    } else {
    	d = DEFAULT_GAIN;
    }
    
    return d;
}


/*
 * ディレクトリごとのGainを保存する
 */
void SaveLocalGain(TCHAR *volumefile, FIL *Fil, double gain){
    FRESULT     res;
    UINT    	br;
    
    res = f_open(Fil, volumefile, FA_READ | FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
		res = f_write(Fil, &gain, sizeof(double), &br);
    	f_close(Fil);
    }
}


/*
 * ASCII文字列のUNICODE文字列化を一時的に作成。
 * 名前をハードコードしているファイルをf_openで開くため。
 */
WCHAR *MakeUnicodeString(const char *s){
	static WCHAR	buf[64];
	int		i;
	int		len;

	len = strlen(s);
	for (i=0; i<len; i++){
		buf[i] = s[i];
	}
	buf[len] = 0;
	return buf;
}


/*
 *	OLED初期化 (pico-examplesのssd1306_i2cから / PAGE ADDRESSING MODEとする)
 */
void SSD1306_init() {
    // Some of these commands are not strictly necessary as the reset
    // process defaults to some of these but they are shown here
    // to demonstrate what the initialization sequence looks like
    // Some configuration values are recommended by the board manufacturer

    uint8_t cmds[] = {
        SSD1306_SET_DISP,               // set display off
        /* memory mapping */
        SSD1306_SET_MEM_MODE,           // set memory address mode 0 = horizontal, 1 = vertical, 2 = page
        0x02,                           // page addressing mode
        //0x00,                           // horizontal addressing mode
        /* resolution and layout */
        SSD1306_SET_DISP_START_LINE,    // set display start line to 0
        SSD1306_SET_SEG_REMAP | 0x01,   // set segment re-map, column address 127 is mapped to SEG0 (default=0)
        SSD1306_SET_MUX_RATIO,          // set multiplex ratio
        SSD1306_HEIGHT - 1,             // Display height - 1
        SSD1306_SET_COM_OUT_DIR | 0x08, // set COM (common) output scan direction. Scan from bottom up, COM[N-1] to COM0
        SSD1306_SET_DISP_OFFSET,        // set display offset
        0x00,                           // no offset
        SSD1306_SET_COM_PIN_CFG,        // set COM (common) pins hardware configuration. Board specific magic number. 
                                        // 0x02 Works for 128x32, 0x12 Possibly works for 128x64. Other options 0x22, 0x32
#if ((SSD1306_WIDTH == 128) && (SSD1306_HEIGHT == 32))
        0x02,                           
#elif ((SSD1306_WIDTH == 128) && (SSD1306_HEIGHT == 64))
        0x12,
#else
        0x02,
#endif
        /* timing and driving scheme */
        SSD1306_SET_DISP_CLK_DIV,       // set display clock divide ratio
        0x80,                           // div ratio of 1, standard freq
        SSD1306_SET_PRECHARGE,          // set pre-charge period
        0xF1,                           // Vcc internally generated on our board
        SSD1306_SET_VCOM_DESEL,         // set VCOMH deselect level
        0x30,                           // 0.83xVcc
        /* display */
        SSD1306_SET_CONTRAST,           // set contrast control
        0xFF,
        SSD1306_SET_ENTIRE_ON,          // set entire display on to follow RAM content
        SSD1306_SET_NORM_DISP,          // set normal (not inverted) display
        SSD1306_SET_CHARGE_PUMP,        // set charge pump
        0x14,                           // Vcc internally generated on our board
        SSD1306_SET_SCROLL | 0x00,      // deactivate horizontal scrolling if set. This is necessary as memory writes will corrupt if scrolling was enabled
        SSD1306_SET_DISP | 0x01, // turn display on
    };

    SSD1306_send_cmd_list(cmds, count_of(cmds));
}


void SSD1306_send_cmd_list(uint8_t *buf, int num) {
    for (int i=0;i<num;i++){
    	//printf("  %d\n", i);
        SSD1306_send_cmd(buf[i]);
    }
}

void SSD1306_send_cmd(uint8_t cmd) {
    // I2C write process expects a control byte followed by data
    // this "data" can be a command or data to follow up a command
    // Co = 1, D/C = 0 => the driver expects a command
    uint8_t buf[2] = {0x80, cmd};
    if (i2c_write_blocking(i2c1, (SSD1306_I2C_ADDR & SSD1306_WRITE_MODE), buf, 2, false) == PICO_ERROR_GENERIC){
    	printf("SSD1306_send_cmd error (%x)\n", cmd);
    }
}

/*
 * PageImageBuff変数に事前に作成されたイメージをOLED全体にまとめて表示する
 */
void DispLcdFullImage(){
	int page;
	//uint8_t	buf[SSD1306_WIDTH + 1];
    absolute_time_t timeout = make_timeout_time_ms(2000);
    
	printf("DispLcdFullImage start\n");
    for (page=0; page <SSD1306_NUM_PAGES; page++){
    	//printf(" page=%d\n", page);
	    SSD1306_send_cmd(0xB0 + page);
    	
    	PageImageBuff[page][0] = 0x40;
    	if (i2c_write_blocking(i2c1, (SSD1306_I2C_ADDR & SSD1306_WRITE_MODE), PageImageBuff[page], SSD1306_WIDTH + 1, false) == PICO_ERROR_GENERIC){
			printf("i2c_write_blocking error (page=%d)\n", page);
    	}
    }
    
	printf("DispLcdFullImage end\n");
}


/*
 *	行を横スクロール表示
 * 	line=0,1,2,3
 */
void DispScrollLine(int line){
	uint8_t	bk;
	int	page;
	
   	for (page=2*line; page<=(2*line + 1); page++){
	    SSD1306_send_cmd(0xB0 + page);
	    // 表示開始バイトの一つ前を0x40に置き換えてから表示し、元に戻す
	    bk = *(PageImageBuff[page] + 1 + ScrollX - 1);
	    *(PageImageBuff[page] + 1 + ScrollX - 1) = 0x40;
		if (i2c_write_blocking(i2c1, (SSD1306_I2C_ADDR & SSD1306_WRITE_MODE), PageImageBuff[page] + 1 + ScrollX - 1, SSD1306_WIDTH + 1, false) == PICO_ERROR_GENERIC){
			printf("i2c_write_blocking error (page=%d)\n", page);
		}
	    *(PageImageBuff[page] + 1 + ScrollX - 1) = bk;
   	}
}


/*
 * 固定文字列の表示
 * 	line=0,1,2,3
 * 	msgno=メッセージNo
 * 	inv=1:反転
 */
void DispMsg(int line, int msgno, int inv, FIL *Fil){
    FRESULT     res;
    UINT        br;
	int			x1;

    printf("DispMsg start %d\n", msgno);
    res = f_open(Fil, MakeUnicodeString(MSGSTRFILE), FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
    	ClearImageBuff();
    	
		res = f_lseek(Fil, (DWORD)(msgno) * MS_IMAGEDATA_LEN);
        if (f_read(Fil, PageImageBuff[2*line] + 1, MS_IMAGEDATAHALF_LEN, &br) != FR_OK){
        	printf("DispMsg f_read_error\n");
        }
        if (f_read(Fil, PageImageBuff[2*line + 1] + 1, MS_IMAGEDATAHALF_LEN, &br) != FR_OK){
        	printf("DispMsg f_read_error\n");
        }
        if (inv == 1){
        	SetMenuItemSelected(line);
        }
        DispLcdFullImage();
    } else {
    	printf("MSGFILE error\n");
    }
    f_close(Fil);
    
    printf("DispMsg end\n");
}


/*
 * PageImageBuff変数に作成されたイメージデータの反転状態を行単位で変更する。
 * line=0,1,2,3
 */
void SetMenuItemSelected(uint8_t line){
	int x1;
	int page;

	page = 2*line;
	for (x1=1; x1<=PageImageWidth[page]; x1++){
		PageImageBuff[page][x1] = ~PageImageBuff[page][x1];
	}
	page = 2*line + 1;
	for (x1=1; x1<=PageImageWidth[page]; x1++){
		PageImageBuff[page][x1] = ~PageImageBuff[page][x1];
	}
}

/*
 *	イメージバッファ初期化
 */
void InitImageBuff(){
	PageImageBuff[0] = Page0Image;
	PageImageBuff[1] = Page1Image;
	PageImageBuff[2] = Page2Image;
	PageImageBuff[3] = Page3Image;
	PageImageBuff[4] = Page4Image;
	PageImageBuff[5] = Page5Image;
	PageImageBuff[6] = Page6Image;
	PageImageBuff[7] = Page7Image;
	
	Page0Image[0] = 0x40;
	Page1Image[0] = 0x40;
	Page2Image[0] = 0x40;
	Page3Image[0] = 0x40;
	Page4Image[0] = 0x40;
	Page5Image[0] = 0x40;
	Page6Image[0] = 0x40;
	Page7Image[0] = 0x40;
	
	PageImageWidth[0] = DL_IMAGEDATAHALF_LEN;
	PageImageWidth[1] = DL_IMAGEDATAHALF_LEN;
	PageImageWidth[2] = SSD1306_WIDTH;
	PageImageWidth[3] = SSD1306_WIDTH;
	PageImageWidth[4] = SSD1306_WIDTH;
	PageImageWidth[5] = SSD1306_WIDTH;
	PageImageWidth[6] = SSD1306_WIDTH;
	PageImageWidth[7] = SSD1306_WIDTH;
}


/*
 *	イメージデータ全クリア
 */
void ClearImageBuff(){
	memset(Page0Image + 1, 0, DL_IMAGEDATAHALF_LEN);
	memset(Page1Image + 1, 0, DL_IMAGEDATAHALF_LEN);
	memset(Page2Image + 1, 0, SSD1306_WIDTH);
	memset(Page3Image + 1, 0, SSD1306_WIDTH);
	memset(Page4Image + 1, 0, SSD1306_WIDTH);
	memset(Page5Image + 1, 0, SSD1306_WIDTH);
	memset(Page6Image + 1, 0, SSD1306_WIDTH);
	memset(Page7Image + 1, 0, SSD1306_WIDTH);

}

/*
 *	特定ページイメージデータクリア
 */
void ClearPageImageBuff(int page){
	memset(PageImageBuff[page] + 1, 0, PageImageWidth[page]);
	PageImageLength[page] = 0;
}


/*
 *	ページごとのイメージデータ
 */
void SetPageImageBuff(int page, uint8_t *buf, int len){
	memcpy(PageImageBuff[page] + 1, buf, len);
	PageImageLength[page] = 0;
	for (int x=PageImageWidth[page]; x>=1; x--){
		if (PageImageBuff[page][x] != 0x00){
			PageImageLength[page] = x;
			break;
		}
	}
}


void error_blink(uint led, int count)
{
    while (true) {
        for (int i = 0; i < count; i++) {
            gpio_put(led, 1);
            sleep_ms(250);
            gpio_put(led, 0);
            sleep_ms(250);
        }
        gpio_put(led, 0);
        sleep_ms(500);
    }
}


